from __future__ import annotations
import asyncio
import json
import logging
from re import S
import socket
import hashlib
import os
import aiosqlite
import secrets
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cli_social.p2p.registry import fetch_and_vrfy_registry
from cli_social.p2p.utils import read_frame, write_frame

logger = logging.getLogger(__name__)

STORE_MAX_MESSAGES = 100
MAX_CONCURRENT_CONN = 500

async def _write_msg(writer: asyncio.StreamWriter, msg: dict) -> None:
    await write_frame(writer, json.dumps(msg).encode())


async def _read_msg(reader: asyncio.StreamReader) -> dict:
    frame = await read_frame(reader)
    return json.loads(frame)


class RelayConnection:
    def __init__(self, peer_id: str, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        self.peer_id = peer_id
        self.reader = reader
        self.writer = writer
        self._alive = True
        self.is_listener: bool = False
        self.msg_count_wind = 0
        self.wind_start = asyncio.get_event_loop().time()
        
    async def send_msg(self, msg: dict) -> None:
        await _write_msg(self.writer, msg)

    async def receive_msg(self) -> dict:
        return await _read_msg(self.reader)

    async def send_raw(self, data: bytes) -> None:
        await write_frame(self.writer, data)

    async def receive_raw(self) -> bytes:
        return await read_frame(self.reader)

    def close(self) -> None:
        self._alive = False
        self.writer.close()

    @property
    def alive(self) -> bool:
        return self._alive


class RelayServer:
    def __init__(self, host: str = "0.0.0.0", port: int = 9100):
        self.host = host
        self.port = port
        host_hash = socket.gethostname()
        self.relay_id = f"relay_{host_hash[:8]}"
        self._online: dict[str, RelayConnection] = {}
        self._db: aiosqlite.Connection | None = None
        self._server: asyncio.Server | None = None
        self._mesh_conns: dict[str, RelayConnection] = {}
        self._mesh_presence: dict[str, str] = {}
        self._mesh_task: asyncio.Task | None = None
        self._me_addr: set[str] = set()
        self._pending_dials: set[str] = set()
        self._pending_acks: dict[str, list[dict]] = {}
        self._trusted_relay_ids: set[str] = set()
        self._ip_conns: dict[str, int] = {}
        self._max_per_ip: int = 5
        self._bg_tasks: set[asyncio.Task] = set()

    async def start(self) -> None:
        db_path = os.getenv("RELAY_DB_PATH", "relay.db")
        self._db = await aiosqlite.connect(db_path)
        await self._db.execute("CREATE TABLE IF NOT EXISTS store (id INTEGER PRIMARY KEY AUTOINCREMENT, peer_id TEXT NOT NULL, payload BLOB NOT NULL, message_id TEXT)")
        try:
            await self._db.execute("ALTER TABLE store ADD COLUMN message_id TEXT")
            await self._db.commit()
        except aiosqlite.OperationalError:
            pass
        await self._db.execute("CREATE TABLE IF NOT EXISTS pending_acks (id INTEGER PRIMARY KEY AUTOINCREMENT, peer_id TEXT NOT NULL, from_id TEXT NOT NULL, message_id TEXT NOT NULL)")
        await self._db.execute("CREATE INDEX IF NOT EXISTS idx_store_peer ON store(peer_id)")
        await self._db.execute("CREATE INDEX IF NOT EXISTS idx_acks_peer ON pending_acks(peer_id)")
        await self._db.commit()
        self._server = await asyncio.start_server(
            self._handle_client,
            self.host,
            self.port
        )
        logger.info(f"Relay listening on {self.host}:{self.port} with ID {self.relay_id}")
        self._mesh_task = asyncio.create_task(self._mesh_mngr())

    async def stop(self) -> None:
        if self._db:
            await self._db.close()
        if self._server:
            self._server.close()
            await self._server.wait_closed()
        for conn in list(self._online.values()) + list(self._mesh_conns.values()):
            conn.close()
        self._online.clear()
        self._mesh_conns.clear()
        logger.info(
            "Relay stopped , You are no longer relaying messages, Please restart when possible , we really need these to keep our app running <3")

    async def run_forever(self) -> None:
        if self._server is None:
            raise RuntimeError("start() needs to called ;-;")
        async with self._server:
            await self._server.serve_forever()
    
    async def _mesh_mngr(self) -> None:
        await asyncio.sleep(5)
        while True:
            try:
                relays = await fetch_and_vrfy_registry(None, accept_community=False)
                self._trusted_relay_ids = {r.get("id") for r in relays if r.get("id")}
                for r in relays:
                    targ_id = r.get("id")
                    addr = r.get("address", "")

                    if targ_id == self.relay_id:
                        continue
                    
                    if addr in self._me_addr:
                        continue

                    if not addr.startswith("tcp://"):
                        continue
                    host, port = addr.replace("tcp://", "").split(":")
                    
                    if targ_id not in self._mesh_conns and targ_id not in self._pending_dials:
                        self._pending_dials.add(targ_id)
                        task = asyncio.create_task(self._dial_mesh_relay(host, int(port) , addr, targ_id))
                        self._bg_tasks.add(task)
                        task.add_done_callback(self._bg_tasks.discard)

            except Exception as e:
                logger.debug(f"Mesh manager loop error: {e}")

            await asyncio.sleep(60)
    
    async def _dial_mesh_relay(self, host: str, port: int, addr_key: str, targ_id: str) -> None:
        conn = None
        try:
            reader, writer = await asyncio.open_connection(host , port)
            reg_msg = {"type": "mesh_register", "relay_id": self.relay_id}
            await write_frame(writer, json.dumps(reg_msg).encode())
            
            ack_raw = await read_frame(reader)
            ack = json.loads(ack_raw)
            remote_relay_id = ack.get("relay_id", addr_key)
            
            if remote_relay_id == self.relay_id:
                logger.info("Its me")
                self._me_addr.add(addr_key)
                writer.close()
                await writer.wait_closed()
                return
            
            if ack.get("type") != "mesh_ok":
                writer.close()
                return
            
            
            conn = RelayConnection(peer_id=addr_key, reader=reader, writer=writer)
            self._mesh_conns[remote_relay_id] = conn
            logger.info(f"established mesh link to {remote_relay_id} | {host}:{port}")
            
            for local_peer in self._online.keys():
                await conn.send_msg({"type": "mesh_add", "peer_id": local_peer, "relay_id": self.relay_id})
                
            while conn.alive:
                msg = await conn.receive_msg()
                await self._handle_mesh_message(msg)
            
        except Exception as e:
            logger.debug(f"failed to connect to {host}:{port}, {e}")
        finally:
            self._pending_dials.discard(targ_id)
            rem_id = locals().get('remote_relay_id')
            if rem_id: 
                for pid, rid in list(self._mesh_presence.items()):
                    if rid == rem_id:
                        self._mesh_presence.pop(pid, None)
            if conn:
                for rid, c in list(self._mesh_conns.items()):
                    if c is conn:
                        self._mesh_conns.pop(rid, None)
                conn.close()
    
    def _broadcast_mesh(self, msg: dict) -> None:
        for conn in self._mesh_conns.values():
            task = asyncio.create_task(conn.send_msg(msg))
            self._bg_tasks.add(task)
            task.add_done_callback(self._bg_tasks.discard)
            
            
    async def _handle_mesh_message(self, msg:dict) -> None:
        msg_type = msg.get("type")
        peer_id = msg.get("peer_id")
        relay_id = msg.get("relay_id")
        
        if msg_type == "mesh_add" and peer_id and relay_id:
            self._mesh_presence[peer_id] = relay_id
            if not self._db:
                return
            async with self._db.execute("SELECT id, payload, message_id FROM store WHERE peer_id = ?", (peer_id,)) as cur:
                stored = await cur.fetchall()
            remote_conn = self._mesh_conns.get(relay_id)
            if stored and remote_conn and remote_conn.alive:
                s_ids = []
                for r_id, p, m_id in stored:
                        wrap = {"type": "mesh_stored_forward", "to": peer_id, "payload": p.hex(), "message_id": m_id}
                        t = asyncio.create_task(remote_conn.send_msg(wrap))
                        self._bg_tasks.add(t)
                        t.add_done_callback(self._bg_tasks.discard)
                        s_ids.append(r_id)
                if s_ids:
                    await self._db.execute(f"DELETE FROM store WHERE id IN ({','.join('?' for _ in s_ids)})", s_ids)
                    await self._db.commit()
            async with self._db.execute("SELECT id, from_id, message_id FROM pending_acks WHERE peer_id = ?", (peer_id,)) as cur:
                acks = await cur.fetchall()
            if acks and remote_conn and remote_conn.alive:
                a_ids = []
                for a_id, frm, m_id in acks:
                    wrap = {"type": "mesh_ack", "to": peer_id, "from": frm, "message_id": m_id}
                    t = asyncio.create_task(remote_conn.send_msg(wrap))
                    self._bg_tasks.add(t)
                    t.add_done_callback(self._bg_tasks.discard)
                    a_ids.append(a_id)
                if a_ids:
                    await self._db.execute(f"DELETE FROM pending_acks WHERE id IN ({','.join('?' for _ in a_ids)})", a_ids)
                    await self._db.commit()
        elif msg_type == "mesh_remove" and peer_id:
            self._mesh_presence.pop(peer_id, None)
        elif msg_type == "mesh_forward":
            target_id = msg.get("to")
            from_id = msg.get("from")
            payload_hex = msg.get("payload")
            msg_id = msg.get("message_id")
            target = self._online.get(target_id)
            if target and target.alive:
                await target.send_msg({"type": "push", "from": from_id, "payload": payload_hex, "message_id": msg_id})
            else:
                self._store_payload(target_id, bytes.fromhex(payload_hex), msg_id=msg_id or "")
        elif msg_type == "mesh_stored_forward":
            target_id = msg.get("to")
            payload_hex = msg.get("payload")
            msg_id = msg.get("message_id")
            target = self._online.get(target_id)
            if target and target.alive:
                wrap = {"type": "stored_message", "payload": payload_hex, "message_id": msg_id}
                t = asyncio.create_task(target.send_msg(wrap))
                self._bg_tasks.add(t)
                t.add_done_callback(self._bg_tasks.discard)
            else:
                self._store_payload(target_id, bytes.fromhex(payload_hex), msg_id=msg_id or "")
        elif msg_type == "mesh_ack":
            target_id = msg.get("to")
            from_id = msg.get("from")
            msg_id = msg.get("message_id")
            target = self._online.get(target_id)
            if target and target.alive:
                await target.send_msg({"type": "delivery_ack", "from": from_id, "message_id": msg_id})
            else:
                self._store_pending_ack(target_id, {"from": from_id, "message_id": msg_id})

    async def _handle_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        peer_addr = writer.get_extra_info("peername")
        
        if len(self._online) >= MAX_CONCURRENT_CONN:
            logger.warning(f"relay at capacity!, dropping conn from {peer_addr}")
            writer.close()
            return
            
        client_ip = peer_addr[0] if peer_addr else "unknown"
        if self._ip_conns.get(client_ip, 0) >= self._max_per_ip:
            logger.warning(f"rate limit, too many connections from {client_ip}")
            writer.close()
            return
        self._ip_conns[client_ip] = self._ip_conns.get(client_ip, 0) + 1
        conn: RelayConnection | None = None
        logger.info(f"new conn from {peer_addr}")

        try:
            msg = await asyncio.wait_for(_read_msg(reader), timeout=10)
            msg_type = msg.get("type")
            
            if msg_type == "mesh_register":
                await self._handle_mesh_reg(msg, reader, writer, peer_addr)
                return
            
            if msg_type != "register" or not msg.get("peer_id") or not msg.get("public_key"):
                await _write_msg(writer, {"type": "error", "reason": "first message must register"})
                return
            peer_id = msg["peer_id"]
            public_key_hex = msg["public_key"]

            try:
                pub_bytes = bytes.fromhex(public_key_hex)
                if hashlib.sha256(pub_bytes).hexdigest() != peer_id:
                    await _write_msg(writer, {"type": "error", "reason": "public key does not match peer id"})
                    return
            except ValueError:
                await _write_msg(writer, {"type": "error", "reason": "invalid public key"})
                return
            
            nonce = secrets.token_hex(32)
            await _write_msg(writer, {"type": "challenge", "nonce": nonce})

            try:
                resp = await asyncio.wait_for(_read_msg(reader), timeout=10)
            except asyncio.TimeoutError:
                await _write_msg(writer, {"type": "error", "reason": "challenge timeout"})
                return
            
            if resp.get("type") != "challenge_response" or not resp.get("signature"):
                await _write_msg(writer, {"type": "error", "reason": "invalid challenge response"})
                return
            
            try:
                ed_pub = Ed25519PublicKey.from_public_bytes(pub_bytes)
                auth_payload = f"cli-sxcl-relay-auth-v1:{nonce}".encode()
                ed_pub.verify(bytes.fromhex(resp["signature"]), auth_payload)
            except Exception:
                await _write_msg(writer, {"type": "error", "reason": "auth failed"})
                return

            conn = RelayConnection(peer_id=msg["peer_id"], reader=reader, writer=writer)
            self._online[conn.peer_id] = conn
            self._broadcast_mesh({"type": "mesh_add", "peer_id": conn.peer_id, "relay_id": self.relay_id})
            await conn.send_msg({"type": "ok"})
            await self._flush_stored(conn)
            await self._flush_pending_acks(conn)

            while conn.alive:
                try:
                    action = await asyncio.wait_for(conn.receive_msg(), timeout=70)
                    action_type = action.get("type")

                    if action_type == "publish":
                        now = asyncio.get_event_loop().time()
                        if now - conn.wind_start > 45:
                            conn.wind_start = now
                            conn.msg_count_wind = 0
                        
                        conn.msg_count_wind += 1
                        if conn.msg_count_wind > 45:
                            logger.warning(f"peer {conn.peer_id[:12]} exceeded rate limit, dropping")
                            await conn.send_msg({"type": "error", "reason": "rate limit exceeded"})
                            conn.close()
                            break
                        await self._handle_publish(conn, action)
                    elif action_type == "client_ack":
                        await self._handle_client_ack(conn, action)
                    elif action_type == "pong":
                        continue
                except asyncio.TimeoutError:
                    await conn.send_msg({"type": "ping"})
        except (asyncio.IncompleteReadError, ConnectionResetError):
            logger.error(f"peer {conn.peer_id[:12] if conn else peer_addr} disconnected")
        finally:
            self._ip_conns[client_ip] = max(0, self._ip_conns.get(client_ip, 0) - 1)
            if conn:
                self._online.pop(conn.peer_id, None)
                self._broadcast_mesh({"type": "mesh_remove", "peer_id": conn.peer_id})
                conn.close()
    
    async def _handle_publish(self, sender: RelayConnection, msg: dict) -> None:
        target_id = msg.get("to")
        payload = msg.get("payload")
        msg_id = msg.get("message_id")

        if not target_id or not payload:
            return
        
        target = self._online.get(target_id)
        if target and target.alive:
            await target.send_msg({"type": "push", "from": sender.peer_id, "payload": payload, "message_id": msg_id})
            await sender.send_msg({"type": "relay_ack", "message_id": msg_id, "status": "relayed"})
            return

        remote_relay_id = self._mesh_presence.get(target_id)
        if remote_relay_id:
            remote_conn = self._mesh_conns.get(remote_relay_id)
            if remote_conn and remote_conn.alive:
                await remote_conn.send_msg({"type": "mesh_forward", "to": target_id, "from": sender.peer_id, "payload": payload, "message_id": msg_id})
                await sender.send_msg({"type": "relay_ack", "message_id": msg_id, "status": "relayed"})
                return
        self._store_payload(target_id, bytes.fromhex(payload), msg_id=msg_id)
        await sender.send_msg({"type": "relay_ack", "message_id": msg_id, "status": "relayed"})
    
    async def _handle_client_ack(self, sender: RelayConnection, msg: dict) -> None:
        target_id = msg.get("to")
        msg_id = msg.get("message_id")
        if not target_id:
            return
        target = self._online.get(target_id)
        if target and target.alive:
            await target.send_msg({"type": "delivery_ack", "from": sender.peer_id, "message_id": msg_id})
        elif target_id in self._mesh_presence:
            remote_relay_id = self._mesh_presence[target_id]
            remote_conn = self._mesh_conns.get(remote_relay_id)
            if remote_conn and remote_conn.alive:
                await remote_conn.send_msg({"type": "mesh_ack", "to": target_id, "from": sender.peer_id, "message_id": msg_id })
            else:
                self._store_pending_ack(target_id, {"from": sender.peer_id, "message_id": msg_id})
        else:
            self._store_pending_ack(target_id, {"from": sender.peer_id, "message_id": msg_id})

    async def _handle_mesh_reg(self, msg: dict, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, addr: tuple) -> None:
        remote_relay_id = msg.get("relay_id")
        if remote_relay_id not in self._trusted_relay_ids:
            logger.warning(f"rejected mesh connection from untrusted relay: {remote_relay_id}")
            await _write_msg(writer, {"type": "error", "reason": "not in trusted registry"})
            writer.close()
            return
        await _write_msg(writer, {"type": "mesh_ok", "relay_id": self.relay_id})

        if remote_relay_id == self.relay_id:
            writer.close()
            return
        
        conn = RelayConnection(peer_id=addr[0], reader=reader, writer=writer)
        self._mesh_conns[remote_relay_id] = conn

        for local_peer in self._online.keys():
            await conn.send_msg({"type": "mesh_add", "peer_id": local_peer, "relay_id": self.relay_id})

        try:
            while conn.alive:
                mesh_msg = await conn.receive_msg()
                await self._handle_mesh_message(mesh_msg)
        except Exception:
            pass
        finally:
            for pid, rid in list(self._mesh_presence.items()):
                if rid == remote_relay_id:
                    self._mesh_presence.pop(pid, None)
            self._mesh_conns.pop(remote_relay_id, None)
            conn.close()
        
    def _store_payload(self, peer_id: str, payload: bytes, msg_id: str = "") -> None:
        async def _save():
            if not self._db:
                return
            await self._db.execute("INSERT INTO store (peer_id, payload, message_id) VALUES (?, ?, ?)", (peer_id, payload, msg_id))
            await self._db.commit()
        t = asyncio.create_task(_save())
        self._bg_tasks.add(t)
        t.add_done_callback(self._bg_tasks.discard)

    def _store_pending_ack(self, peer_id: str, ack_data: dict) -> None:
        async def _save():
            if not self._db:
                return
            await self._db.execute("INSERT INTO pending_acks (peer_id, from_id, message_id) VALUES (?, ?, ?)", (peer_id, ack_data["from"], ack_data["message_id"]))
            await self._db.commit()
        t = asyncio.create_task(_save())
        self._bg_tasks.add(t)
        t.add_done_callback(self._bg_tasks.discard)
    
    async def _flush_pending_acks(self, conn: RelayConnection) -> None:
        if not self._db:
            return
        peer_id = conn.peer_id
        async with self._db.execute("SELECT id, from_id, message_id FROM pending_acks WHERE peer_id = ?", (peer_id,)) as cur:
            acks = await cur.fetchall()
        if acks:
            logger.debug(f"flushing {len(acks)} acks to {peer_id[:12]}")
            s_ids = []
            for a_id, frm, m_id in acks:
                try:
                    await conn.send_msg({"type": "delivery_ack", "from": frm, "message_id": m_id})
                    s_ids.append(a_id)
                except Exception as e:
                    logger.error(f"failed to flush ack to {peer_id[:12]}, {e}")
                    break
            if s_ids:
                await self._db.execute(f"DELETE FROM pending_acks WHERE id IN ({','.join('?' for _ in s_ids)})", s_ids)
                await self._db.commit()
    
    def _unregister(self, conn: RelayConnection) -> None:
        self._online.pop(conn.peer_id, None)
        conn.close()
        logger.info(f"peer {conn.peer_id[:12]} unregisterd")

    async def _flush_stored(self, conn: RelayConnection) -> None:
        if not self._db:
            return
        peer_id = conn.peer_id
        async with self._db.execute("SELECT id, payload, message_id FROM store WHERE peer_id = ?", (peer_id,)) as cur:
            stored = await cur.fetchall()
        if stored:
            logger.debug(f"flushing {len(stored)} stored to {peer_id[:12]}")
            s_ids = []
            for r_id, p, m_id in stored:
                try:
                    await conn.send_msg({"type": "stored_message", "payload": p.hex(), "message_id": m_id})
                    s_ids.append(r_id)
                except Exception as e:
                    logger.error(f"failed to flush msg to {peer_id[:12]}, {e}")
                    break
            if s_ids:
                await self._db.execute(f"DELETE FROM store WHERE id IN ({','.join('?' for _ in s_ids)})", s_ids)
                await self._db.commit()
    

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="0.0.0.0")
    parser.add_argument("--port", type=int, default=9100)
    args = parser.parse_args()
    # I need max logging ahhhh
    
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(levelname)s %(name)s %(message)s", handlers=[logging.StreamHandler(), logging.FileHandler("relay.log")])
    
    async def _main():
        server = RelayServer(host=args.host, port=args.port)
        await server.start()
        logger.info(f"relay running on {args.host}:{args.port}")
        await server.run_forever()
    asyncio.run(_main())


from __future__ import annotations
import asyncio
import hashlib
import json
import logging
from datetime import datetime, timezone
from dataclasses import dataclass
from typing import Optional
from kademlia.network import Server
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

logger = logging.getLogger(__name__)

DEFAULT_DHT_PORT = 63012
@dataclass
class PeerInfo:
    peer_id: str
    username: str = ""
    last_seen: str = ""
    noise_pubkey_hex: str = ""
    home_relay: str = ""
    ed25519_pubkey_hex: str = ""
    signature: str = ""
    
    def _get_signable_data(self) -> bytes:
        d = f"{self.peer_id}:{self.username}:{self.last_seen}:{self.noise_pubkey_hex}:{self.home_relay}:{self.ed25519_pubkey_hex}"
        return d.encode()

    def to_json(self) -> str:
        return json.dumps({
            "peer_id": self.peer_id,
            "username": self.username,
            "last_seen": self.last_seen,
            "noise_pubkey_hex": self.noise_pubkey_hex,
            "home_relay": self.home_relay,
            "ed25519_pubkey_hex": self.ed25519_pubkey_hex,
            "signature": self.signature
        })
    @classmethod
    def from_json(cls, data:str) -> "PeerInfo":
        d = json.loads(data)
        return cls(
            peer_id=d["peer_id"],
            username=d.get("username", ""),
            last_seen=d.get("last_seen", ""),
            noise_pubkey_hex=d.get("noise_pubkey_hex", ""),
            home_relay=d.get("home_relay", ""),
            ed25519_pubkey_hex=d.get("ed25519_pubkey_hex", ""),
            signature=d.get("signature", "")
        )
    
    @property
    def fingerprint(self) -> str:
        if not self.noise_pubkey_hex:
            return ""
        import hashlib
        return hashlib.sha256(bytes.fromhex(self.noise_pubkey_hex)).hexdigest()
    
    def is_valid(self) -> bool:
        if not self.ed25519_pubkey_hex or not self.signature:
            return False
        try:
            pub_bytes = bytes.fromhex(self.ed25519_pubkey_hex)
            if hashlib.sha256(pub_bytes).hexdigest() != self.peer_id:
                return False
            pub_key = Ed25519PublicKey.from_public_bytes(pub_bytes)
            pub_key.verify(bytes.fromhex(self.signature), self._get_signable_data())
            return True
        except Exception:
            return False

class DHTNode:
    def __init__(
        self,
        peer_id: str,
        host: str = "0.0.0.0",
        port: int = DEFAULT_DHT_PORT,
        bootstrap_nodes: list[tuple[str, int]] | None = None,
        ed25519_private_key=None
    ):
        self.peer_id = peer_id
        self.host = host
        self.port = port
        self.bootstrap_nodes = bootstrap_nodes or []
        self.ed25519_private_key = ed25519_private_key
        self._server = Server(ksize=8, alpha=5)
        self._started = False
    
    async def start(self) -> None:
        await self._server.listen(self.port)
        self._server.protocol.wait_timeout = 2.0
        self._started = True
        logger.info(f"TNT?? nah DHT node listening on port {self.port}")
        
        if self.bootstrap_nodes:
            asyncio.create_task(self._server.bootstrap(self.bootstrap_nodes))
            logger.info(f"bootsrapped with {self.bootstrap_nodes}")
        else:
            logger.warning("No nodes configured | in local mode")
    
    async def reannounce(self, username: str = "", noise_pubkey_hex: str = "", home_relay: str = "") -> None:
        while self._started:
            await self.announce(username=username, noise_pubkey_hex=noise_pubkey_hex, home_relay=home_relay)
            await asyncio.sleep(60)
    
    async def stop(self) -> None:
        if self._started:
            self._server.stop()
            self._started = False
            logger.info("DHT node stopped")
            
    async def announce(self, username: str = "", noise_pubkey_hex: str = "", home_relay: str = "") -> None:
        info = PeerInfo(
            peer_id=self.peer_id,
            username=username,
            last_seen=datetime.now(timezone.utc).isoformat(),
            noise_pubkey_hex=noise_pubkey_hex,
            home_relay=home_relay
        )
        if self.ed25519_private_key:
            pub_bytes = self.ed25519_private_key.public_key().public_bytes(Encoding.Raw, PublicFormat.Raw)
            info.ed25519_pubkey_hex = pub_bytes.hex()
            info.signature = self.ed25519_private_key.sign(info._get_signable_data()).hex()
        await self._server.set(self.peer_id, info.to_json())
        logger.info(f"Announced self to DHT {self.peer_id[:12]} on relay {home_relay}")
        
    async def lookup(self, peer_id: str) -> Optional[PeerInfo]:
        result = await self._server.get(peer_id)
        if result is None:
            logger.debug(f"peer {peer_id[:12]} not found")
            return None
        try:
            info = PeerInfo.from_json(result)
            if not info.is_valid():
                logger.error(f"DHT signature verification failed for {peer_id[:12]} it is unsigned or invalid signature")
                return None
            return info
        except Exception as e:
            logger.error(f"parsing failed for peer info {peer_id[:12]} {e}")
            return None

    async def is_online(self, peer_id: str, threshold_seconds: int = 90) -> bool:
        info = await self.lookup(peer_id)
        if not info or not info.last_seen:
            return False
        last_seen = datetime.fromisoformat(info.last_seen)
        age = (datetime.now(timezone.utc) - last_seen).total_seconds()
        return age < threshold_seconds
    
    async def get_value(self, key: str) -> str | None:
        return await self._server.get(key)
    
    async def set_value(self, key: str, value: str) -> None:
        await self._server.set(key, value)
        
    async def __aenter__(self):
        await self.start()
        return self
    
    async def __aexit__(self, *_):
        await self.stop()



# I thought for reusability , I should put those functions here and import them in other files

from __future__ import annotations
import asyncio
import struct

MAX_FRAME_SIZE = 64 * 1024

async def read_frame(reader: asyncio.StreamReader) -> bytes:
    header = await reader.readexactly(4)
    length = struct.unpack(">I", header)[0]
    if length > MAX_FRAME_SIZE:
        raise ValueError(f"frame is large: {length} bytes")
    return await reader.readexactly(length)

async def write_frame(writer: asyncio.StreamWriter, data: bytes) -> None:
    writer.write(struct.pack(">I", len(data)) + data)
    await writer.drain()


from __future__ import annotations
import hashlib
import json
import os
import secrets
import platformdirs
from pathlib import Path
from argon2.low_level import hash_secret_raw, Type
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey
)
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.serialization import (
    Encoding, 
    NoEncryption,
    PrivateFormat,
    PublicFormat
)

DEFAULT_KEY_FILE = Path(platformdirs.user_config_dir("cli-social")) / "identity.enc"
ARGON2_TIME_COST = 3
ARGON2_MEM_COST = 65536
ARGON2_PARALLELISM = 2
ARGON2_HASH_LEN = 32
ARGON2_SALT_LEN = 16
AES_NONCE_LEN = 12


def _derive_key(passphrase: str, salt: bytes) -> bytes:
    return hash_secret_raw(
        secret=passphrase.encode(),
        salt=salt,
        time_cost=ARGON2_TIME_COST,
        memory_cost=ARGON2_MEM_COST,
        parallelism=ARGON2_PARALLELISM,
        hash_len=ARGON2_HASH_LEN,
        type=Type.ID,
    )
    
def _peer_id_from_public_key(public_key: Ed25519PublicKey) -> str:
    raw = public_key.public_bytes(Encoding.Raw, PublicFormat.Raw)
    return hashlib.sha256(raw).hexdigest()

def _key_file_for(data_dir: Path | None) -> Path:
    if data_dir is not None:
        return data_dir / "identity.enc"
    return DEFAULT_KEY_FILE

def generate_identity(
    passphrase: str,
    username: str = "",
    key_file: Path = DEFAULT_KEY_FILE
) -> str:

    if key_file.exists():
        raise FileExistsError(f"Identity file already exists: {key_file}")
    
    private_key = Ed25519PrivateKey.generate()
    public_key = private_key.public_key()
    private_bytes = private_key.private_bytes(Encoding.Raw, PrivateFormat.Raw, NoEncryption())
    
    noise_private = X25519PrivateKey.generate()
    noise_private_bytes = noise_private.private_bytes(Encoding.Raw, PrivateFormat.Raw, NoEncryption())
    noise_public_bytes = noise_private.public_key().public_bytes(Encoding.Raw, PublicFormat.Raw)
    
    salt = secrets.token_bytes(ARGON2_SALT_LEN)
    aes_key = _derive_key(passphrase, salt)
    nonce = secrets.token_bytes(AES_NONCE_LEN)
    aesgcm = AESGCM(aes_key)
    ciphertext= aesgcm.encrypt(nonce, private_bytes, None)
    noise_nonce = secrets.token_bytes(AES_NONCE_LEN)
    noise_ciphertext = aesgcm.encrypt(noise_nonce, noise_private_bytes, None)
    
    peer_id = _peer_id_from_public_key(public_key)
    payload = {
        "version": 1,
        "argon2_salt": salt.hex(),
        "aes_nonce": nonce.hex(),
        "ciphertext": ciphertext.hex(),
        "noise_nonce": noise_nonce.hex(),
        "noise_ciphertext": noise_ciphertext.hex(),
        "noise_public_key": noise_public_bytes.hex(),
        "peer_id": peer_id,
        "username": username,
    }
    
    key_file.parent.mkdir(parents=True, exist_ok=True)
    key_file.write_text(json.dumps(payload, indent=2))
    
    try:
        os.chmod(key_file, 0o600)
    except NotImplementedError:
        pass    # I hate you windows user :)
    
    return peer_id

def load_identity(
    passphrase: str,
    key_file: Path = DEFAULT_KEY_FILE
) -> tuple[Ed25519PrivateKey, Ed25519PublicKey, str, str, bytes]:
    
    if not key_file.exists():
        raise FileNotFoundError(f"No identity found at {key_file}")
    
    payload = json.loads(key_file.read_text())
    if payload.get("version") !=1:
        raise ValueError(f"Unsupported identity file version: {payload.get('version')}")
    
    salt = bytes.fromhex(payload["argon2_salt"])
    nonce = bytes.fromhex(payload["aes_nonce"])
    ciphertext = bytes.fromhex(payload["ciphertext"])
    aes_key = _derive_key(passphrase, salt)
    aesgcm = AESGCM(aes_key)
    
    try:
        private_bytes = aesgcm.decrypt(nonce, ciphertext, None)
    except Exception as exc:
        raise ValueError("Decryption failed!! wrong passphrase or corrupt key file") from exc
    
    noise_nonce = bytes.fromhex(payload["noise_nonce"])
    noise_ciphertext = bytes.fromhex(payload["noise_ciphertext"])
    noise_private_bytes = aesgcm.decrypt(noise_nonce, noise_ciphertext, None)
    
    
    private_key = Ed25519PrivateKey.from_private_bytes(private_bytes)
    public_key = private_key.public_key()
    peer_id = _peer_id_from_public_key(public_key)
    username = payload.get("username", "")
    
    return private_key, public_key, peer_id, username, noise_private_bytes

def identity_exists(key_file: Path = DEFAULT_KEY_FILE) -> bool:
    return key_file.exists()

def sign_registry_doc(private_key: Ed25519PrivateKey, registry_data: dict) -> dict:
    payload = {k: v for k, v in registry_data.items() if k != "signatures"}
    message = json.dumps(payload, sort_keys=True).encode('utf-8')
    signature = private_key.sign(message)
    pub_hex = private_key.public_key().public_bytes(Encoding.Raw, PublicFormat.Raw).hex()
    registry_data["signatures"] = { pub_hex: signature.hex() }
    return registry_data


from __future__ import annotations
import asyncio
import json
import logging
import time
import traceback
from collections import deque
from pathlib import Path
from typing import Callable, Awaitable
from cli_social.p2p.transport import encrypt_blob
from cli_social.p2p.dht import DHTNode
from cli_social.storage import Storage, DEFAULT_DB_PATH
from cli_social.p2p.utils import read_frame, write_frame
from cli_social.p2p.registry import fetch_and_vrfy_registry
from cli_social.p2p.transport import decrypt_blob
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

logger = logging.getLogger(__name__)


MessageCallback = Callable[[str, str], Awaitable[None]]
StatusCallback = Callable[[str, str], Awaitable[None]]

class Daemon:
    def __init__(
        self,
        peer_id: str,
        private_key: bytes,
        username: str = "",
        dht_port: int = 63012,
        bootstrap_nodes: list[tuple[str, int]] | None = None,
        on_message: MessageCallback | None = None,
        status_callback: StatusCallback | None = None,
        db_path: Path | None = None,
        relay_host: str | None = None,
        relay_port: int | None = None,
        cached_relay: dict | None = None,
        ed25519_private_key=None
    ):
        self.peer_id = peer_id
        self.private_key = private_key
        self.username = username
        self.dht_port = dht_port
        self.bootstrap_nodes = bootstrap_nodes or []
        self.on_message = on_message
        self.status_callback = status_callback
        self.db_path = db_path
        self.relay_host = relay_host
        self.relay_port = relay_port
        self.cached_relay = cached_relay
        self.ed25519_private_key = ed25519_private_key
        self._dht: DHTNode | None = None
        self._storage: Storage | None = None
        self._running = False
        self._relay_reader: asyncio.StreamReader | None = None
        self._relay_writer: asyncio.StreamWriter | None = None
        self._relay_task: asyncio.Task | None = None
        self._bg_tasks: set[asyncio.Task] = set()
        self._recent_message_ids = deque(maxlen=200)
        self.noise_pubkey_hex = ""
        


    # coding this at 1am , I need caffeine !!
    async def start(self) -> None:
        noise_priv_key = X25519PrivateKey.from_private_bytes(self.private_key)
        self.noise_pubkey_hex = (
            noise_priv_key.public_key()
            .public_bytes(Encoding.Raw, PublicFormat.Raw)
            .hex()
        )
        self._storage = await Storage.open(self.db_path or DEFAULT_DB_PATH, encryption_key=self.private_key)
        self._dht = DHTNode(
            peer_id=self.peer_id,
            port=self.dht_port,
            bootstrap_nodes=self.bootstrap_nodes,
            ed25519_private_key=self.ed25519_private_key
        )
        await self._dht.start()
        self._running = True
        if self.relay_host:
            await self._connect_to_relay()
        else:
            await self._discover_and_connect_relay()
            
        t_mngr = asyncio.create_task(self._run_relay_mngr())
        self._bg_tasks.add(t_mngr)
        t_mngr.add_done_callback(self._bg_tasks.discard)
        hm_str = f"{self.relay_host}:{self.relay_port}" if self.relay_host else ""
        await self._dht.announce(
            username=self.username,
            noise_pubkey_hex=self.noise_pubkey_hex,
            home_relay=hm_str,
        )
        
    async def _on_message(self, peer_id: str, content: str, client_message_id: str | None = None) -> None:
        if client_message_id:
            if client_message_id in self._recent_message_ids:
                return
            self._recent_message_ids.append(client_message_id)
            if self._storage and await self._storage.is_msg_dup(client_message_id):
                return
        if self._storage:
            await self._storage.save_message(peer_id=peer_id, sender_peer_id=peer_id, content=content, is_outgoing=False)
        if self.on_message:
            await self.on_message(peer_id, content)

    # Idk how but on every night coding session I end up in this file , fr bruh, also HI reviewer I know you are going to read this comment, can you please do me a favor to dm me on slack @speedhawks , pls ?? uwu :3

    async def _run_relay_mngr(self):
            is_first_run = True
            while self._running:
                if not self._relay_writer:
                    if self.relay_host:
                        await self._connect_to_relay()
                    elif is_first_run and self.cached_relay:
                        self.relay_host = self.cached_relay["host"]
                        self.relay_port = self.cached_relay["port"]
                        try:
                            await asyncio.wait_for(self._connect_to_relay(), timeout=3)
                        except Exception:
                            self.relay_host = None
                            self.cached_relay = None
                    else:
                        await self._discover_and_connect_relay()

                if self._relay_writer:
                    is_first_run = False
                    logger.info("announcing presence to DHT")
                    home_relay_str = f"{self.relay_host}:{self.relay_port}"
                    await self._dht.announce(
                        username=self.username,
                        noise_pubkey_hex=self.noise_pubkey_hex,
                        home_relay=home_relay_str,
                    )
                    await asyncio.sleep(60)
                else:
                    await asyncio.sleep(30)
    
    async def send_chat_msg(self, target_peer_id: str, content: str, client_msg_id: str) -> None:
        if not self._relay_writer:
            raise ConnectionError("Not connected to relay network")
        
        pubkey_hex = ""
        if self._storage:
            pubkey_hex = await self._storage.get_contact_pubkey(target_peer_id)
        
        if not pubkey_hex and self._dht:
            peer_info = await self._dht.lookup(target_peer_id)
            if peer_info:
                pubkey_hex = peer_info.noise_pubkey_hex
                if self._storage:
                    await self._storage.update_contact_pubkey(target_peer_id, pubkey_hex)
                if peer_info.fingerprint:
                    await self._storage.update_contact_fingerprint(target_peer_id, peer_info.fingerprint)
        if not pubkey_hex:
            raise ValueError("target public key unknown")
        
        encrypted_blob = await encrypt_blob(self.peer_id, self.private_key, bytes.fromhex(pubkey_hex), content, client_msg_id, ed25519_private_key=self.ed25519_private_key)
        msg = {"type": "publish", "to": target_peer_id, "payload": encrypted_blob.hex(), "message_id": client_msg_id}
        await write_frame(self._relay_writer, json.dumps(msg).encode())
        logger.debug(f"pushed packet {client_msg_id[:8]} to relay for {target_peer_id}")

    async def _connect_to_relay(self) -> None:
        try:
            reader, writer = await asyncio.open_connection(
                self.relay_host, self.relay_port
            )
            reg = {"type": "register", "peer_id": self.peer_id}
            if self.ed25519_private_key:
                pub_bytes = self.ed25519_private_key.public_key().public_bytes(Encoding.Raw, PublicFormat.Raw)
                reg["public_key"] = pub_bytes.hex()
            await write_frame(writer, json.dumps(reg).encode())
            resp = json.loads(await read_frame(reader))
            if resp.get("type") == "challenge" and self.ed25519_private_key:
                nonce = resp["nonce"]
                auth_payload = f"cli-sxcl-relay-auth-v1:{nonce}".encode()
                signature = self.ed25519_private_key.sign(auth_payload)
                await write_frame(writer, json.dumps({"type": "challenge_response", "signature": signature.hex(),}).encode())            
                resp = json.loads(await read_frame(reader))
            if resp.get("type") != "ok":
                logger.error(f"relay register failed {resp.get('reason')}")
                writer.close()
                return
            self._relay_reader = reader
            self._relay_writer = writer
            logger.info(f"registered with relay at {self.relay_host}:{self.relay_port}")
            self._relay_task = asyncio.create_task(
                self._listen_relay(reader, writer), name="relay listener"
            )
        except Exception as e:
            logger.error(f"could not connect {e}")

    async def _listen_relay(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        try:
            while self._running:
                msg = json.loads(await read_frame(reader))
                msg_type = msg.get("type")

                if msg_type == "push":
                    sender_id = msg.get("from")
                    payload_hex = msg.get("payload")
                    client_msg_id = msg.get("message_id")

                    try:
                        decrypted = await decrypt_blob(self.private_key, bytes.fromhex(payload_hex))
                        verified_sender = decrypted.get("peer_id", sender_id)
                        await self._on_message(verified_sender, decrypted["content"], client_msg_id)
                        ack = {"type": "client_ack", "to": sender_id, "message_id": client_msg_id}
                        await write_frame(writer, json.dumps(ack).encode())
                    except Exception as e:
                        logger.error(f"failed to process push, {e}")
                elif msg_type == "relay_ack":
                    msg_id = msg.get("message_id")
                    if self._storage:
                        await self._storage.mark_relayed(msg_id)
                    if self.status_callback:
                        await self.status_callback(msg_id, "relayed")
                elif msg_type == "delivery_ack":
                    msg_id = msg.get("message_id")
                    if self._storage:
                        await self._storage.mark_delivered(msg_id)
                    if self.status_callback:
                        await self.status_callback(msg_id, "delivered")                      
                elif msg_type == "stored_message":
                    payload_hex = msg.get("payload")
                    client_msg_id = msg.get("message_id")
                    try:
                        decrypted = await decrypt_blob(self.private_key, bytes.fromhex(payload_hex))
                        sender_id = decrypted.get("peer_id", "unknown")
                        if not client_msg_id:
                            client_msg_id = decrypted.get("client_message_id")
                        await self._on_message(peer_id=sender_id, content=decrypted["content"], client_message_id=client_msg_id)
                        if client_msg_id:
                            ack = {"type": "client_ack", "to": sender_id, "message_id": client_msg_id}
                            await write_frame(writer, json.dumps(ack).encode())
                    except Exception as e:
                        logger.error(f"failed to decrypt stored msg, {e}")
                elif msg_type == "ping":
                    logger.debug("relay keepalive ping")
                    await write_frame(writer, json.dumps({"type": "pong"}).encode())
                elif msg_type == "error":
                    logger.warning(f"relay error {msg.get('reason')}")

        except asyncio.IncompleteReadError:
            logger.info("relay closed conn")
        except Exception as e:
            logger.error(f"relay listener error {e}")
        finally:
            self._relay_reader = None
            self._relay_writer = None
            logger.info(f"relay listener stopped at: {traceback.format_exc()}")
            if self._running:
                await asyncio.sleep(3)
                await self._connect_to_relay()

    async def run_forever(self) -> None:
        while self._running:
            await asyncio.sleep(1)

    async def stop(self) -> None:
        self._running = False
        if self._relay_task:
            self._relay_task.cancel()
        for task in list(self._bg_tasks):
            task.cancel()
        if self._relay_writer:
            self._relay_writer.close()
        if self._dht:
            await self._dht.stop()
        if self._storage:
            await self._storage.close()
        logger.info("dameon stopped")


    async def _discover_and_connect_relay(self) -> None:

        if self._dht is None:
            logger.error("DHT not running, cannot dicsover relays")
            return

        dht_payload = await self._dht.get_value("relays.v1")

        try:
            relays = await fetch_and_vrfy_registry(dht_payload, accept_community=False)
            if not relays:
                logger.error("No relays found in the verified registry")
                return

            tcp_relays = [
                r for r in relays if r.get("address", "").startswith("tcp://")
            ]
            if not tcp_relays:
                logger.error("No valid relays found")
                return

            logger.info(f"found {len(tcp_relays)}, measuring latency..")

            async def measure_latency(relay: dict) -> tuple[dict, float]:
                addr = relay["address"].replace("tcp://", "")
                host, port = addr.split(":")
                start_time = time.perf_counter()
                try:
                    reader, writer = await asyncio.wait_for(
                        asyncio.open_connection(host, int(port)), timeout=3
                    )
                    writer.close()
                    await writer.wait_closed()
                    latency = time.perf_counter() - start_time
                    return relay, latency
                except Exception:
                    return relay, float("inf")

            ping_tasks = [measure_latency(r) for r in tcp_relays]
            results = await asyncio.gather(*ping_tasks)
            alive_relays = [(r, lat) for r, lat in results if lat != float("inf")]

            if not alive_relays:
                logger.error("all relays unreacsble at the moment, please retry")
                return

            alive_relays.sort(key=lambda x: x[1])
            for r, lat in alive_relays:
                region = r.get("meta", {}).get("region", "unknown")
                logger.info(f"  {r['id']} ({region}), {lat * 100:.0f} ms")

            chosen = alive_relays[0][0]

            addr = chosen["address"].replace("tcp://", "")
            host, port = addr.split(":")

            self.relay_host = host
            self.relay_port = int(port)
            logger.info(
                f"Chose fasetest relay from registry {self.relay_host}:{self.relay_port}"
            )

            await self._connect_to_relay()

        except Exception as e:
            logger.error(f"relay sequence failed {e}")



from __future__ import annotations
import asyncio
from pathlib import Path
import click
from cli_social.identity import (
    generate_identity,
    load_identity,
    identity_exists,
    _key_file_for
)
import json
import time
import platformdirs
import logging
from cli_social.storage import db_path_for
from cli_social.p2p.daemon import Daemon
from cli_social.p2p.dht import DHTNode
from cli_social.p2p.relay import RelayServer
from cli_social.identity import sign_registry_doc
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

DEFAULT_BOOTSTRAP_NODES = [
    ("64.225.108.241", 52342),
    ("139.59.77.182", 51560),
    ("45.55.209.227", 52585),
    ("138.197.209.223", 51311)
]


DATA_DIR_OPTION = click.option(
    "--data-dir",
    default=None,
    type=click.Path(file_okay=False, path_type=Path),
    help="db path",
    envvar="SXCL_DATA_DIR"
)

def _require_identity(data_dir: Path | None) -> tuple[str, bytes, str, Ed25519PublicKey, Ed25519PrivateKey]:
    key_file = _key_file_for(data_dir)
    if not identity_exists(key_file):
        click.echo("\n Welcome to CLI-SXCL! No identity found")
        do_setup = click.prompt("Would you like to set one up now? [Y/n]", default="y")
        if do_setup.lower().startswith('y'):
            init.callback(data_dir)
            click.echo("\n---Setup Complete! ---\n")
        else:
            click.echo("Okay! Run `sxcl init` when you are ready")
            raise SystemExit(1)
    
    passphrase = click.prompt("Enter your passphrase", hide_input=True)
    
    try:
        private_key, public_key, peer_id, username, noise_private_bytes = load_identity(passphrase, key_file)
        return peer_id, noise_private_bytes, username, public_key, private_key
    except ValueError:
        click.echo("Wrong passphrase! you forgot your password??")
        raise SystemExit(1)
    
@click.group()
def main():
    pass

@main.group()
def registry():
    pass

@registry.command()
@click.argument("input_file", type=click.File('r'))
@click.argument("output_file", type=click.File('w'))
@DATA_DIR_OPTION
def sign(input_file, output_file, data_dir):
    key_file = _key_file_for(data_dir)
    if not identity_exists(key_file):
        click.echo("No identity found. Run `sxcl init` first.")
        raise SystemExit(1)
    
    passphrase = click.prompt("Enter passphrase for signing key", hide_input=True)
    try:
        private_key, _, _, _, _ = load_identity(passphrase, key_file)
    except ValueError:
        click.echo("Wrong passphrase! you forgot your password?? dumbahh")
        raise SystemExit(1)
    
    registry_data = json.load(input_file)
    registry_data["version"] = registry_data.get("version", 0) + 1
    registry_data["timestamp"] = int(time.time())
    registry_data["signatures"] = {}
    signed_registry = sign_registry_doc(private_key, registry_data)
    json.dump(signed_registry, output_file, indent=2)
    click.echo(f"signed registry wrote it to {output_file.name}")

@main.command()
@DATA_DIR_OPTION
def init(data_dir):
    key_file = _key_file_for(data_dir)
    key_path = str(key_file).replace(str(Path.home()), "~")
    if identity_exists(key_file):
        click.echo(f"Identity already exists at {key_file}")
        click.echo("Delete it mannually if you want to start new (you will lose your account tho, be careful)")
        return

    click.echo("Creating a new identity..\n")
    username = click.prompt("Username (this is optional, press enter to skip)", default="")
    
    while True:
        passphrase = click.prompt("Passphrase", hide_input=True)
        if len(passphrase) < 8:
            click.echo("Passphrase must be at least 8 characters, (more efforts)")
            continue
        confirm = click.prompt("Confirm Passphrase", hide_input=True)
        if passphrase != confirm:
            click.echo("Passphrase don't match, try again")
            continue
        break
    
    click.echo("\n Deriving key for you (this takes a few seconds) be patient.")
    peer_id = generate_identity(passphrase, username, key_file)
    
    click.echo("\n Identity created")
    click.echo(f"   Username : {username or '(none)'}")
    click.echo(f"   Peer ID  : {peer_id}")
    click.echo(f"   Key File : {key_path}")
    click.echo("\n Remember your passphrase, there is no recovery")

@main.command()
@DATA_DIR_OPTION
def nuke(data_dir):
    key_file = _key_file_for(data_dir)
    db_path = db_path_for(data_dir)

    click.secho("\n WARNING: Thid will permanently delete your identity and ALL messages!", fg="red", bold=True)
    confirm = click.prompt("Are you ABSOLUTELY sure? Type 'DELETE' to confirm", default="")

    if confirm == "DELETE":
        if key_file.exists():
            key_file.unlink()
            click.echo(f"[-] Deleted identity at {key_file}")
        if db_path.exists():
            db_path.unlink()
            click.echo(f"[-] Deleted database at {db_path}")
        
        state_path = db_path.parent / "state.json"
        if state_path.exists():
            state_path.unlink()
        
        click.secho("\nWipe complete. You are a ghost again!", fg="green", bold=True)
    else:
        click.echo("\nWipe cancelled. Your identity and data remain intact")

    
@main.command()
@DATA_DIR_OPTION
def whoami(data_dir):
    peer_id, _, username, public_key, _ = _require_identity(data_dir)
    pub_key_hex = public_key.public_bytes(Encoding.Raw, PublicFormat.Raw).hex()
    click.echo(f"   Username : {username or '(none)'}")
    click.echo(f"   Peer ID  : {peer_id}")
    click.echo(f"   Pub Key  : {pub_key_hex}")

@main.command()
@DATA_DIR_OPTION
@click.option("--dht-port", default=6969, help="DHT listen port")
@click.option("--bootstrap", multiple=True, help="bootstrap nodes as host:port")
@click.option("--relay", default=None, help="relay node as host:port")
def tui(dht_port, bootstrap, relay, data_dir):

    log_dir = Path(platformdirs.user_log_dir("cli-social"))
    log_dir.mkdir(parents=True, exist_ok=True)

    logging.basicConfig(
    filename=str(Path(data_dir or log_dir) / "debug.log"),
    level=logging.DEBUG,
    format="%(asctime)s %(name)s %(levelname)s %(message)s"
    )
    
    from cli_social.tui import run
    peer_id, private_key, username, _, ed25519_key = _require_identity(data_dir)    
    
    bootstrap_nodes = []
    if bootstrap:
        for node in bootstrap:
            h, p = node.rsplit(":", 1)
            bootstrap_nodes.append((h, int(p)))
    else:
        bootstrap_nodes = DEFAULT_BOOTSTRAP_NODES
        
    relay_host, relay_port = None, 9100
    if relay:
        rh, rp = relay.rsplit(":", 1)
        relay_host, relay_port = rh, int(rp)
        
    run(peer_id=peer_id, private_key=private_key, username=username, dht_port=dht_port, bootstrap_nodes=bootstrap_nodes, db_path=db_path_for(data_dir), relay_host=relay_host, relay_port=relay_port, ed25519_key=ed25519_key)

@main.command()
@DATA_DIR_OPTION
@click.option("--dht-port", default=6969, help="DHT listen port")
@click.option("--bootstrap", multiple=True, help="Bootstrap nodes as host:port")
def daemon(dht_port, bootstrap, data_dir):
    peer_id, private_key, username, _, ed25519_key = _require_identity(data_dir)
    
    bootstrap_nodes = []
    if bootstrap:
        for node in bootstrap:
            h, p = node.rsplit(":", 1)
            bootstrap_nodes.append((h, int(p)))
    else:
        bootstrap_nodes = DEFAULT_BOOTSTRAP_NODES
        
    async def _run():
        d = Daemon(
            peer_id=peer_id,
            private_key=private_key,
            username=username,
            dht_port=dht_port,
            bootstrap_nodes=bootstrap_nodes,
            db_path=db_path_for(data_dir),
            ed25519_private_key=ed25519_key
        )
        await d.start()
        click.echo(f"DHT port {dht_port}")
        click.echo(f"Peer ID {peer_id[:16]}....{peer_id[-8:]}")
        click.echo("Press Ctrl+C to stop")
        try:
            await d.run_forever()
        except (KeyboardInterrupt, asyncio.CancelledError):
            pass
        finally:
            await d.stop()
            click.echo("\n Daemon stopped")
            
    asyncio.run(_run())
  
@main.command()
@click.option("--host", default="0.0.0.0", help="the host to listen on")
@click.option("--relay-port", default=19853, help="tcp port for relay server")
@click.option("--dht-port", default=6969, help="UDP port for the dht node")
@click.option("--relay", is_flag=True, help="run as a message relay node")
@click.option("--bootstrap", is_flag=True, help="run as dht bootstrap node")
def node(host, relay_port, dht_port, relay, bootstrap):
    if not (relay or bootstrap):
        click.echo("You must specify role (--relay, --boostrap)")
        return
    
    import logging
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(name)s %(levelname)s %(message)s")
    
    async def _run_node():
        tasks = []
        if bootstrap:
            dht = DHTNode(peer_id="bootstrap_node", host=host, port=dht_port, bootstrap_nodes=DEFAULT_BOOTSTRAP_NODES)
            tasks.append(dht.start())
            click.echo(f"DHT bootstrap node listening on UDP {host}:{dht_port}")
            
        if relay:
            relay_server = RelayServer(host=host, port=relay_port)
            tasks.append(asyncio.create_task(relay_server.start()))
            click.echo(f"Relay node listening on TCP {host}:{relay_port}")
        
        # if this crashes
        if tasks:
            await asyncio.gather(*tasks)
            await asyncio.Future()
        else:
            click.echo("No tasks to run")
    
    try:
        asyncio.run(_run_node())
    except KeyboardInterrupt:
        click.echo("\nShutting down node!") 
                   


from __future__ import annotations
import json
import logging
import hashlib
from datetime import datetime, timezone
from typing import Callable, Awaitable
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from noise.connection import NoiseConnection, Keypair

logger = logging.getLogger(__name__)


MessageHandler = Callable[[str, str, int], Awaitable[None]]
ReceiptHandler = Callable[[int], Awaitable[None]]


class PeerOfflineError(ConnectionError):
    def __init__(self, message, writer) -> None:
        super().__init__(message)
        self.writer = writer

async def encrypt_blob(our_peer_id: str, our_private_key: bytes, their_pubkey: bytes, content: str, client_message_id: str, ed25519_private_key=None):
    noise = NoiseConnection.from_name(b"Noise_X_25519_ChaChaPoly_SHA256")
    noise.set_as_initiator()
    noise.set_keypair_from_private_bytes(Keypair.STATIC, our_private_key)
    noise.set_keypair_from_public_bytes(Keypair.REMOTE_STATIC, their_pubkey)
    noise.start_handshake()
    inner = {
            "type": "message",
            "peer_id": our_peer_id,
            "content": content,
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "client_message_id": client_message_id
    }
    if ed25519_private_key:
        inner_bytes = json.dumps(inner, separators=(',', ':'), sort_keys=True).encode()
        inner["signature"] = ed25519_private_key.sign(inner_bytes).hex()
        inner["ed25519_pubkey"] = ed25519_private_key.public_key().public_bytes(Encoding.Raw, PublicFormat.Raw).hex()
    payload = json.dumps(inner).encode()
    encrypted_message = noise.write_message(payload=payload)
    return bytes(encrypted_message)

async def decrypt_blob(our_private_key: bytes, encrypted_blob: bytes) -> dict:
    noise = NoiseConnection.from_name(b"Noise_X_25519_ChaChaPoly_SHA256")
    noise.set_as_responder()
    noise.set_keypair_from_private_bytes(Keypair.STATIC, our_private_key)
    noise.start_handshake()
    
    decrypted_payload = noise.read_message(encrypted_blob)
    data = json.loads(decrypted_payload)
    if "signature" in data and "ed25519_pubkey" in data:
        pubkey_hex = data.pop("ed25519_pubkey")
        signature_hex = data.pop("signature")
        pub_bytes = bytes.fromhex(pubkey_hex)
        expected_peer_id = hashlib.sha256(pub_bytes).hexdigest()
        if expected_peer_id != data.get("peer_id"):
            raise ValueError("sender pubkey does not match claimed peer id (SPOOFING ALERT!)")
        ed_pub = Ed25519PublicKey.from_public_bytes(pub_bytes)
        verify_data = json.dumps({k: v for k, v in data.items()}, separators=(',', ':'), sort_keys=True).encode()
        ed_pub.verify(bytes.fromhex(signature_hex), verify_data)
        data["verified"] = True
    else:
        raise ValueError("message has no signature, rejecting unsigned paayload")
    return data




import json
import urllib.error
import urllib.request
import asyncio
import logging
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.exceptions import InvalidSignature

logger = logging.getLogger(__name__)

TRUSTED_PUBKEYS = ["60c01e07546f068b051fe54a2f0b0e34a67a129b6e9eccb235923706734def9c"]
REGISTRY_URL = "https://raw.githubusercontent.com/kusuta012/cli-social/main/registry.signed.json"

def vrfy_registry_sign(public_key_hex: str, data: dict, sign_hex: str) -> bool:
    try:
        public_key_bytes = bytes.fromhex(public_key_hex)
        public_key = Ed25519PublicKey.from_public_bytes(public_key_bytes)
        sign_bytes = bytes.fromhex(sign_hex)
        
        payload_to_vrfy = {k: v for k, v in data.items() if k != "signatures"}
        
        message = json.dumps(payload_to_vrfy, sort_keys=True).encode('utf-8')
        
        public_key.verify(sign_bytes, message)
        return True
    except (InvalidSignature, ValueError) as e:
        logger.debug(f"Signature verification failed: {e}")
        return False

async def fetch_and_vrfy_registry(dht_payload: str | None, accept_community: bool = False) -> list[dict]:
    registry_data = None
    
    if dht_payload:
        try:
            registry_data = json.loads(dht_payload)
            logger.info("loaded registry from DHT")
        except json.JSONDecodeError:
            logger.warning("not a valid json received from DHT")
    
    if not registry_data:
        logger.info("fetching registry from github")    
        def _fetch():
            req = urllib.request.Request(REGISTRY_URL, headers={'User-Agent': 'cli-social-node/1.0'})
            with urllib.request.urlopen(req, timeout=10) as resp:
                return json.loads(resp.read().decode('utf-8'))
        
        try:
            registry_data = await asyncio.to_thread(_fetch)
        except Exception as e:
            logger.error(f"failed to fetch registry from {REGISTRY_URL}: {e}")
            raise
    
    signatures = registry_data.get("signatures", {})
    valid_sigs = 0
    
    for pubkey, sig in signatures.items():
        if pubkey in TRUSTED_PUBKEYS or accept_community:
            if vrfy_registry_sign(pubkey, registry_data, sig):
                valid_sigs+=1
    
    if valid_sigs == 0 and not accept_community:
        raise ValueError("Registery rejected, no valid trusted signatures found")


    return registry_data.get("relays",[])



from __future__ import annotations
import asyncio
import hashlib
import logging
import uuid
import json
from datetime import datetime, timezone
from typing import Optional
from pathlib import Path
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import Header, Footer, Static, Input, ListView, ListItem, Label
from textual.containers import Horizontal, Vertical, ScrollableContainer
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.message import Message
from textual import on
from cli_social.storage import Storage, DEFAULT_DB_PATH
from cli_social.p2p.daemon import Daemon


logger = logging.getLogger(__name__)


class ConversationItem(ListItem):
    def __init__(self, peer_id: str, username: str, unread: int = 0) -> None:
        super().__init__()
        self.peer_id = peer_id
        self.username = username
        self.unread = unread

    def compose(self) -> ComposeResult:
        unread_badge = f" [bold red]({self.unread})[/bold red]" if self.unread else ""
        name = self.username or self.peer_id[:12] + "..."
        yield Label(f" [bold #babbf1]{name}[/bold #babbf1]{unread_badge}", classes="convo-label")


class Sidebar(Vertical):
    DEFAULT_CSS = """
    Sidebar {
        width: 30;
        border-right: vkey $primary;
        background: $panel;
    }
    #sidebar-title {
        padding: 1 2;
        text-style: bold;
        background: $boost;
        width: 100%;
        color: $text;
        border-bottom: solid $primary;
    }
    .convo-label {
        padding: 1 2;
    }
    """

    def compose(self) -> ComposeResult:
        yield Label(" Conversations", id="sidebar-title")
        yield ListView(id="conversation-list")


class MessageBubble(Static):
    DEFAULT_CSS = """
    MessageBubble {
        padding: 1 4;
        margin: 0;
        width: 100%;
        layout: vertical;
    }
    MessageBubble.outgoing {
        align: right top;
    }
    MessageBubble.incoming {
        align: left top;
    }
    .msg-meta {
        color: $text-muted;
        text-style: bold italic;
        width: auto;
    }
    .msg-content {
        width: auto;
        max-width: 80%;
    }
    MessageBubble.outgoing .msg-content {
        color: $success;
    }
    MessageBubble.incoming .msg-content {
        color: $text;
    }
    """

    def __init__(
        self,
        content: str,
        sender: str,
        sent_at: str,
        is_outgoing: bool,
        delivered: int = 0,
        message_id: int = -1,
        client_message_id: str = "",
        **kwargs,
    ):
        super().__init__()
        self.msg_content = content
        self.sender = sender
        self.sent_at = sent_at
        self.is_outgoing = is_outgoing
        self.delivered = delivered
        self.message_id = message_id
        self.client_message_id = client_message_id
        self.status = {0: "pending", 1: "relayed", 2: "delivered"}.get(
            delivered, "pending"
        )
        self.add_class("outgoing" if is_outgoing else "incoming")

    def compose(self) -> ComposeResult:
        time_str = self._local_time_str()
        prefix = "You" if self.is_outgoing else self.sender[:12]
        icon = ""
        if self.is_outgoing:
            icon = {
                "pending": "[red]●[/red]",
                "relayed": "[yellow]●[/yellow]",
                "delivered": "[green]✓[/green]",
                "failed": "[bold red]![/bold red]",
            }.get(self.status, " ")
        yield Label(f"{prefix} {time_str} {icon}", classes="msg-meta")
        yield Label(self.msg_content, classes="msg-content")

    def _local_time_str(self) -> str:
        try:
            dt = datetime.fromisoformat(self.sent_at).astimezone()
            return dt.strftime("%H:%M")
        except (ValueError, TypeError):
            return self.sent_at[11:16]

    def update_status(self, status: str) -> None:
        self.status = status
        icon = {
            "pending": "[red]●[/red]",
            "relayed": "[yellow]●[/yellow]",
            "delivered": "[green]✓[/green]",
            "failed": "[bold red]![/bold red]",
        }.get(status, " ")
        time_str = self._local_time_str()
        prefix = "You" if self.is_outgoing else self.sender[:12]
        meta = self.query_one(".msg-meta", Label)
        meta.update(f"{prefix} {time_str} {icon}")


class ChatPane(Vertical):
    DEFAULT_CSS = """
    ChatPane {  
        width: 1fr;
        background: $surface;
    }

    #chat-header {
        padding: 1 2;
        background: $boost;
        text-style: bold;
        border-bottom: solid $primary;
        width: 100%;
        content-align: left middle;
    }
    #message-scroll {
        height: 1fr;
        padding: 1 2;
        overflow-y: auto;
    }
    .ascii-banner {
        width: 100%;
        height: 100%;
        content-align: center middle;
    }
    """

    current_peer_id: reactive[Optional[str]] = reactive(None)
    current_username: str = ""
    current_fingerprint: str = ""

    def compose(self) -> ComposeResult:
        yield Label("Select a conversation", id="chat-header")
        scroll = ScrollableContainer(id="message-scroll")
        banner = Static(
            r"""
                                                                                
                                                                                
                                                                                
   ▒████▒  ██         ██████              ▒████▒   ██▓  ▓██    ▒████▒  ██       
  ▓██████  ██         ██████             ▒██████   ▒██  ██▒   ▓██████  ██       
 ▒██▒  ░█  ██           ██               ██▒  ▒█    ██▓▓██   ▒██▒  ░█  ██       
 ██▓       ██           ██               ██         ▒████▒   ██▓       ██       
 ██░       ██           ██               ███▒        ████    ██░       ██       
 ██        ██           ██               ▒█████▒     ▒██▒    ██        ██       
 ██        ██           ██                ░█████▒    ▒██▒    ██        ██       
 ██░       ██           ██      █████        ▒███    ████    ██░       ██       
 ██▓       ██           ██      █████          ██   ▒████▒   ██▓       ██       
 ▒██▒  ░█  ██           ██               █▒░  ▒██   ██▒▒██   ▒██▒  ░█  ██       
  ▓██████  ████████   ██████             ███████▒  ▒██  ██▒   ▓██████  ████████ 
   ▒████▒  ████████   ██████             ░█████▒   ██▓  ▓██    ▒████▒  ████████ 
                                                                                
                                                                                
                                                                                
                                                                                    
    """,
            classes="ascii-banner",
        )

        with scroll:
            banner.styles.width = "100%"
            banner.styles.height = "100%"
            banner.styles.padding = (4, 8)
            banner.styles.color = "#f38ba8"
            yield banner

    async def close_chat(self) -> None:
        self.current_peer_id = None
        self.current_username = ""
        self.current_fingerprint = ""
        self.query_one("#chat-header", Label).update("Select a conversation")

        scroll = self.query_one("#message-scroll", ScrollableContainer)
        await scroll.remove_children()

        banner = Static(
            r"""
                                                                                
                                                                                
                                                                                
   ▒████▒  ██         ██████              ▒████▒   ██▓  ▓██    ▒████▒  ██       
  ▓██████  ██         ██████             ▒██████   ▒██  ██▒   ▓██████  ██       
 ▒██▒  ░█  ██           ██               ██▒  ▒█    ██▓▓██   ▒██▒  ░█  ██       
 ██▓       ██           ██               ██         ▒████▒   ██▓       ██       
 ██░       ██           ██               ███▒        ████    ██░       ██       
 ██        ██           ██               ▒█████▒     ▒██▒    ██        ██       
 ██        ██           ██                ░█████▒    ▒██▒    ██        ██       
 ██░       ██           ██      █████        ▒███    ████    ██░       ██       
 ██▓       ██           ██      █████          ██   ▒████▒   ██▓       ██       
 ▒██▒  ░█  ██           ██               █▒░  ▒██   ██▒▒██   ▒██▒  ░█  ██       
  ▓██████  ████████   ██████             ███████▒  ▒██  ██▒   ▓██████  ████████ 
   ▒████▒  ████████   ██████             ░█████▒   ██▓  ▓██    ▒████▒  ████████ 
                                                                                
                                                                                
                                                                                
                                                                                                                  
            """,
            classes="ascii-banner",
        )
        banner.styles.content_align = ("center", "middle")
        banner.styles.color = "#f38ba8"
        banner.styles.height = "100%"

        await scroll.mount(banner)

    async def load_messages(
        self, peer_id: str, username: str) -> None:
        self.current_peer_id = peer_id
        self.current_username = username
        self.current_fingerprint = ""
        self.query_one("#chat-header", Label).update(
            f"{username or peer_id[:16]}  [dim]●[/dim]"
        )

        scroll = self.query_one("#message-scroll", ScrollableContainer)
        await scroll.remove_children()

        async with await self.app.get_db() as s:
            messages = await s.get_messages(peer_id, limit=50)

        for msg in messages:
            bubble = MessageBubble(
                content=msg["content"],
                sender=username,
                sent_at=msg["sent_at"],
                is_outgoing=bool(msg["is_outgoing"]),
                delivered=msg["delivered"],
                message_id=msg["id"],
                client_message_id=msg["client_message_id"],
            )
            await scroll.mount(bubble)
        scroll.scroll_end(animate=False)

    async def append_message(
        self,
        content: str,
        sender: str,
        sent_at: str,
        is_outgoing: bool,
        delivered: bool = False,
        message_id: int = -1,
        client_message_id: str = "",
    ) -> MessageBubble:
        scroll = self.query_one("#message-scroll", ScrollableContainer)
        bubble = MessageBubble(
            content=content,
            sender=sender,
            sent_at=sent_at,
            is_outgoing=is_outgoing,
            delivered=delivered,
            message_id=message_id,
            client_message_id=client_message_id,
        )
        await scroll.mount(bubble)
        scroll.scroll_end(animate=True)
        return bubble

    def set_status(self, status: str) -> None:
        dot = {
            "ok": "[green]●[/green]",
            "error": "[red]●[/red]",
            "connecting": "[yellow]●[/yellow]",
        }.get(status, "[dim]●[/dim]")
        header = self.query_one("#chat-header", Label)
        name = self.current_username or (self.current_peer_id or "")[:16]
        fp_str = ""
        if self.current_fingerprint:
            fp = self.current_fingerprint
            fp_str = f" [bold #8c8fa1](Fingerprint: {fp[:6]})[/bold #8c8fa1]"  # hate magenta :3
        header.update(f"[bold #babbf1]{name}[/bold #babbf1]{fp_str} {dot}")

class InputBar(Horizontal):
    DEFAULT_CSS = """
    InputBar {
        height: 3;
        border-top: solid $primary;
        background: $boost;
    }
    #message-input {
        width: 1fr;
        border: none;
        background: transparent;
    }
    #message-input:focus {
        border: none;
    }
    """

    def compose(self) -> ComposeResult:
        yield Input(placeholder="Type a message... (Enter to send)", id="message-input")


class ChatModal(ModalScreen):
    DEFAULT_CSS = """
    ChatModal {
        align: center middle;
        background: $background 50%;
    }
    #modal-dialog {
        width: 60;
        height: auto;
        padding: 2 4;
        background: $surface;
        border: thick $primary;
    }
    #chat-modal-error {
        margin-top: 1;
        content-align: center middle;
    }
    """
    BINDINGS = [("escape", "dismiss", "Close")]

    def compose(self) -> ComposeResult:
        with Vertical(id="modal-dialog"):
            yield Label("Start a new conversation")
            yield Label("Enter Peer ID: ")
            yield Input(placeholder="Peer ID...", id="chat-input")
            yield Label("", id="chat-modal-error")

    def _validate_peer_id(self, peer_id: str, own_peer_id: str) -> str | None:
        if len(peer_id) != 64 or not all(c in "0123456789abcdef" for c in peer_id):
            return "Invalid peer ID, must be 64 hex characters"
        if peer_id == own_peer_id:
            return "That's your own peer ID, FOOL"
        return None

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        peer_id = event.value.strip().lower()
        app: CLISocialApp = self.app
        err_label = self.query_one("#chat-modal-error", Label)
        error = self._validate_peer_id(peer_id, app.peer_id)
        if error:
            err_label.update(f"[red]{error}[/red]")
            return

        async with await app.get_db() as s:
            convos = await s.get_conversations()
        existing = [c for c in convos if c["peer_id"] == peer_id]
        if existing:
            err_label.update("[yellow]Conversation already exists[/yellow]")
            await app.on_existing_conversation(peer_id)
            self.dismiss()
            return

        await self.app.start_conversation(peer_id)
        self.dismiss()


class CLISocialApp(App):
    TITLE = "cli-sxcl"
    SUB_TITLE = "decentralized encrypted messaging"
    CSS = """
    Screen {
        layout: vertical;
    }
    
    #main-area {
        height: 1fr;
    }
    
    #sidebar-title {
        padding: 1;
        text-style: bold;
        background: $boost;
    }
    """

    BINDINGS = [
        ("ctrl+q", "quit", "Quit"),
        ("ctrl+n", "new_chat", "New Chat"),
        Binding("ctrl+d", "close_chat", "Close Chat", priority=True),
        ("ctrl+b", "toggle_sidebar", "Toggle Sidebar"),
        ("escape", "blur_input", "Sidebar"),
    ]

    def __init__(
        self,
        peer_id: str,
        private_key: bytes,
        username: str,
        dht_port: int = 63012,
        bootstrap_nodes: list[tuple[str, int]] | None = None,
        db_path: Path = DEFAULT_DB_PATH,
        relay_host: str | None = None,
        relay_port: int = 9100,
        ed25519_key=None,
    ) -> None:
        super().__init__()
        self.peer_id = peer_id
        self.private_key = private_key
        self.username = username
        self.dht_port = dht_port
        self.bootstrap_nodes = bootstrap_nodes or []
        self.db_path = db_path
        self._daemon: Daemon | None = None
        self._daemon_task: asyncio.Task | None = None
        self._presence_task: asyncio.Task | None = None
        self.relay_host = relay_host
        self.relay_port = relay_port
        self.ed25519_key = ed25519_key
        self.state_path = db_path.parent / "state.json"
        self.cached_relay = None
        if self.state_path.exists():
            try:
                self.cached_relay = json.loads(self.state_path.read_text())
                logger.info(f"loaded cache relay {self.cached_relay}")
            except Exception:
                logger.warning("could not load cached relay")
                pass

    def compose(self) -> ComposeResult:
        yield Header()
        with Horizontal(id="main-area"):
            yield Sidebar()
            yield ChatPane()
        yield InputBar()
        yield Footer()

    async def get_db(self) -> Storage:
        return await Storage.open(self.db_path, encryption_key=self.private_key)

    async def on_mount(self) -> None:
        self.sub_title = f"Peer ID: {self.peer_id}"
        self.run_worker(self._start_daemon)
        await self._load_conversations()
        self.set_focus(self.query_one("#conversation-list"))

    async def _start_daemon(self) -> None:
        self.notify(
            "Starting daemon, this may take a while",
            title="System",
            severity="information",
            timeout=5,
        )
        self.query_one(ChatPane).set_status("connecting")
        try:
            self._daemon = Daemon(
                peer_id=self.peer_id,
                private_key=self.private_key,
                username=self.username,
                bootstrap_nodes=self.bootstrap_nodes,
                on_message=self.handle_incoming,
                status_callback=self.handle_status_update,
                db_path=self.db_path,
                relay_host=self.relay_host,
                relay_port=self.relay_port,
                cached_relay=self.cached_relay,
                ed25519_private_key=self.ed25519_key,
            )
            await self._daemon.start()
            self._daemon_task = asyncio.create_task(
                self._daemon.run_forever(), name="daemon"
            )
            if self._daemon._dht:
                self._presence_task = asyncio.create_task(
                    self._presence_refresh(), name="presence_refresh"
                )
            self.notify(
                f"Daemon up! | DHT port {self.dht_port}",
                severity="information",
                timeout=3,
            )
        except Exception as e:
            self.notify(f"Daemon failed to start: {e}", severity="error", timeout=8)
            self.query_one(ChatPane).set_status("error")

    async def _load_conversations(self) -> None:
        async with await self.get_db() as s:
            convos = await s.get_conversations()

        lv = self.query_one("#conversation-list", ListView)
        await lv.clear()
        for c in convos:
            await lv.append(
                ConversationItem(
                    peer_id=c["peer_id"],
                    username=c.get("username") or "",
                    unread=c.get("unread_count") or 0,
                )
            )

    async def start_conversation(self, peer_id: str) -> None:
        async with await self.get_db() as s:
            await s.get_or_create_conversation(peer_id)
            if self._daemon and self._daemon._dht:
                info = await self._daemon._dht.lookup(peer_id)
                if info:
                    await s.add_contact(
                        peer_id,
                        username=info.username,
                        public_key=info.noise_pubkey_hex,
                        fingerprint=info.fingerprint,
                    )
        await self._load_conversations()

    async def on_existing_conversation(self, peer_id: str) -> None:
        lv = self.query_one("#conversation-list", ListView)
        items = list(lv.query(ConversationItem))
        for i, item in enumerate(items):
            if item.peer_id == peer_id:
                lv.index = i
                chat = self.query_one(ChatPane)
                await chat.load_messages(
                    item.peer_id, item.username
                )
                self.set_focus(self.query_one("#message-input"))

    async def on_list_view_selected(self, event: ListView.Selected) -> None:
        item = event.item
        if not isinstance(item, ConversationItem):
            return
        chat = self.query_one(ChatPane)

        await chat.load_messages(item.peer_id, item.username)

        async with await self.get_db() as s:
            convo_id = await s.get_or_create_conversation(item.peer_id)
            await s.mark_read(convo_id)
            contact = await s.get_contact(item.peer_id)
            item.unread = 0
            badge_label = item.query_one(".convo-label", Label)
            name = item.username or item.peer_id[:12] + "..."
            badge_label.update(f" [bold #babbf1]{name}[/bold #babbf1]")

        if contact and contact.get("fingerprint"):
            my_noise_hex = self._daemon.noise_pubkey_hex if self._daemon else ""
            if my_noise_hex:
                my_fp = hashlib.sha256(bytes.fromhex(my_noise_hex)).hexdigest()
                their_fp = contact["fingerprint"]
                combined = "".join(sorted([my_fp, their_fp]))
                chat.current_fingerprint = hashlib.sha256(combined.encode()).hexdigest()
        chat.set_status("connecting")
        if self._daemon and self._daemon._dht:

            async def fetch_presence():
                online = await self._daemon._dht.is_online(item.peer_id)
                if chat.current_peer_id == item.peer_id:
                    chat.set_status("ok" if online else "error")

            asyncio.create_task(fetch_presence())
        self.set_focus(self.query_one("#message-input"))

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        if event.input.id != "message-input":
            return
        content = event.value.strip()
        if not content:
            return

        chat = self.query_one(ChatPane)
        if not chat.current_peer_id:
            self.notify("Select a conversation first!")
            return

        event.input.value = ""
        await self._send_message(chat.current_peer_id, content)

    async def _send_message(self, peer_id: str, content: str) -> None:
        chat = self.query_one(ChatPane)
        client_msg_id = str(uuid.uuid4())
        now = datetime.now(timezone.utc).isoformat()
        async with await self.get_db() as s:
            msg_id = await s.save_message(
                peer_id=peer_id,
                sender_peer_id=self.peer_id,
                content=content,
                is_outgoing=True,
                client_message_id=client_msg_id,
            )

        await chat.append_message(
            content,
            "You",
            now,
            True,
            message_id=msg_id,
            client_message_id=client_msg_id,
        )
        try:
            await self._daemon.send_chat_msg(peer_id, content, client_msg_id)
        except Exception as e:
            self.notify(f"Network faliure: {e}", title="System", severity="error")
            for b in self.query(MessageBubble):
                if b.client_message_id == client_msg_id:
                    b.update_status("failed")

    async def handle_incoming(self, peer_id: str, content: str) -> None:
        now = datetime.now(timezone.utc).isoformat()
        chat = self.query_one(ChatPane)
        username = ""
        async with await self.get_db() as s:
            contact = await s.get_contact(peer_id)
            if contact and contact.get("username"):
                username = contact["username"]
            elif self._daemon and self._daemon._dht:
                info = await self._daemon._dht.lookup(peer_id)
                if info:
                    if info.username:
                        username = info.username
                    await s.add_contact(
                        peer_id,
                        username=info.username,
                        public_key=info.noise_pubkey_hex,
                        fingerprint=info.fingerprint,
                    )
        display_name = username or peer_id[:12]
        if chat.current_peer_id == peer_id:
            self.call_next(chat.append_message, content, display_name, now, False)
        else:
            self.notify(f"New message from {display_name}")
        self.call_next(self._load_conversations)

    async def handle_status_update(self, msg_id: str, status: str) -> None:
        self.call_next(self._apply_status, msg_id, status)

    def _apply_status(self, msg_id: str, status: str) -> None:
        for b in self.query(MessageBubble):
            if b.client_message_id == msg_id:
                b.update_status(status)
                break

    async def _presence_refresh(self) -> None:
        while True:
            await asyncio.sleep(60)
            try:
                chat = self.query_one(ChatPane)
                if chat.current_peer_id and self._daemon and self._daemon._dht:
                    online = await self._daemon._dht.is_online(chat.current_peer_id)
                    chat.set_status("ok" if online else "error")
            except Exception:
                pass

    async def action_quit(self) -> None:
        if self._daemon and self._daemon.relay_host:
            relay_info = {
                "host": self._daemon.relay_host,
                "port": self._daemon.relay_port,
            }
            self.state_path.write_text(json.dumps(relay_info))
        if self._daemon:
            await self._daemon.stop()
        if self._daemon_task:
            self._daemon_task.cancel()
        if self._presence_task:
            self._presence_task.cancel()
        self.exit()

    def action_new_chat(self) -> None:
        self.push_screen(ChatModal())

    def action_blur_input(self) -> None:
        self.set_focus(self.query_one("#conversation-list"))

    def action_toggle_sidebar(self) -> None:
        sidebar = self.query_one(Sidebar)
        sidebar.display = not sidebar.display

    async def action_close_chat(self) -> None:
        chat = self.query_one(ChatPane)
        await chat.close_chat()


def run(
    peer_id: str = "",
    private_key: bytes = b"",
    username: str = "",
    dht_port: int = 63012,
    bootstrap_nodes: list[tuple[str, int]] | None = None,
    db_path: Path = DEFAULT_DB_PATH,
    relay_host: str | None = None,
    relay_port: int = 9100,
    ed25519_key=None,
) -> None:
    CLISocialApp(
        peer_id=peer_id,
        private_key=private_key,
        username=username,
        dht_port=dht_port,
        bootstrap_nodes=bootstrap_nodes or [],
        db_path=db_path,
        relay_host=relay_host,
        relay_port=relay_port,
        ed25519_key=ed25519_key,
    ).run()



from __future__ import annotations

import hashlib
import logging
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

import aiosqlite
import platformdirs
import base64
from cryptography.fernet import Fernet, InvalidToken

DEFAULT_DB_PATH = Path(platformdirs.user_data_dir("cli-social")) / "data.db"

SCHEMA = """
CREATE TABLE IF NOT EXISTS contacts (
    peer_id TEXT PRIMARY KEY,
    username TEXT NOT NULL DEFAULT '',
    public_key TEXT NOT NULL DEFAULT '',
    host TEXT NOT NULL DEFAULT '',
    port INTEGER NOT NULL DEFAULT 0,
    added_at TEXT NOT NULL,
    fingerprint TEXT NOT NULL DEFAULT ''
);

CREATE TABLE IF NOT EXISTS conversations (
    id  INTEGER PRIMARY KEY AUTOINCREMENT,
    peer_id TEXT NOT NULL UNIQUE,
    last_message_at TEXT,
    unread_count INTEGER NOT NULL DEFAULT 0
);

CREATE TABLE IF NOT EXISTS messages (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    conversation_id INTEGER NOT NULL REFERENCES conversations(id),
    sender_peer_id TEXT NOT NULL,
    client_message_id TEXT,
    content TEXT NOT NULL,
    sent_at TEXT NOT NULL,
    is_outgoing INTEGER NOT NULL DEFAULT 0,
    delivered INTEGER NOT NULL DEFAULT 0
);

CREATE INDEX IF NOT EXISTS idx_messages_conversation ON messages(conversation_id);
CREATE INDEX IF NOT EXISTS idx_messages_client_id ON messages(client_message_id);
CREATE INDEX IF NOT EXISTS idx_conversations_peer ON conversations(peer_id);
"""

def db_path_for(data_dir: Path | None) -> Path:
    if data_dir is not None:
        return data_dir / "data.db"
    return DEFAULT_DB_PATH

def _now() -> str:
    return datetime.now(timezone.utc).isoformat()

class Storage:
    def __init__(self, db: aiosqlite.Connection, fernet: Fernet | None = None):
        self._db = db
        self._fernet = fernet
        
    @classmethod
    async def open(cls, db_path: Path = DEFAULT_DB_PATH, encryption_key: bytes | None = None) -> "Storage":
        fernet = None
        if encryption_key:
            digest = hashlib.sha256(encryption_key).digest()
            key_b64 = base64.urlsafe_b64encode(digest)
            fernet = Fernet(key_b64)
        db_path.parent.mkdir(parents=True, exist_ok=True)
        db = await aiosqlite.connect(db_path)
        db.row_factory = aiosqlite.Row
        await db.executescript(SCHEMA)
        async with db.execute("PRAGMA table_info(messages)") as cursor:
            columns = [row["name"] for row in await cursor.fetchall()]
            if "client_message_id" not in columns:
                try:
                    await db.execute("ALTER TABLE messages ADD column client_message_id TEXT")
                except aiosqlite.OperationalError:
                    logging.error("no problem pookie")
        await db.commit()
        return cls(db, fernet=fernet)
    
    async def close(self):
        await self._db.close()
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, *_):
        await self.close()
        
    async def add_contact(
        self,
        peer_id: str,      
        username: str = "",
        public_key: str = "",
        host: str = "",
        port: int = 0,
        fingerprint: str = ""
    ) -> None:
        await self._db.execute(
            """
            INSERT INTO contacts (peer_id, username, public_key, host, port, added_at, fingerprint)
            VALUES (?, ?, ?, ?, ?, ?, ?)
            ON CONFLICT(peer_id) DO UPDATE SET
                username = excluded.username,
                public_key = excluded.public_key,
                host = excluded.host,
                port = excluded.port,
                fingerprint = excluded.fingerprint
            """,
            (peer_id, username, public_key, host, port, _now(), fingerprint)
        )
        await self._db.commit()
        
    async def get_contacts(self) -> list[dict]:
        async with self._db.execute(
            "SELECT peer_id, username, public_key, host, port, added_at, fingerprint FROM contacts ORDER BY username"
        ) as cursor:
            rows = await cursor.fetchall()
            return [dict(row) for row in rows]
    
    async def get_contact(self, peer_id: str) -> Optional[dict]:
        async with self._db.execute(
            "SELECT peer_id, username, public_key, added_at, fingerprint FROM contacts WHERE peer_id = ?",
            (peer_id,),
        ) as cursor:
            row = await cursor.fetchone()
            return dict(row) if row else None
    
    async def get_contact_pubkey(self, peer_id: str) -> str | None:
        async with self._db.execute("SELECT public_key FROM contacts WHERE peer_id = ?", (peer_id,)) as cursor:
            row = await cursor.fetchone()
            if not row:
                return None
            return row["public_key"] or None
    
    async def update_contact_pubkey(self, peer_id: str, pubkey_hex: str) -> None:
        await self._db.execute("UPDATE contacts SET public_key = ? WHERE peer_id = ?", (pubkey_hex, peer_id))
        await self._db.commit()
    
    async def update_contact_fingerprint(self, peer_id:str, fingerprint: str) -> None:
        await self._db.execute("UPDATE contacts SET fingerprint = ? WHERE peer_id = ?", (fingerprint, peer_id))
        await self._db.commit()
               
    async def get_or_create_conversation(self, peer_id: str) -> int:
        async with self._db.execute(
            "SELECT id FROM conversations WHERE peer_id = ?", (peer_id,)
        ) as cursor:
            row = await cursor.fetchone()
            if row:
                return row["id"]
        
        cursor = await self._db.execute(
            "INSERT INTO conversations (peer_id, last_message_at, unread_count) VALUES (?, ?, 0)",
            (peer_id, _now()),
        )
        await self._db.commit()
        assert cursor.lastrowid is not None, "insert failed no lastrowid returned"
        return cursor.lastrowid
    
    async def get_conversations(self) -> list[dict]:
        async with self._db.execute(
            """
            SELECT c.id, c.peer_id, c.last_message_at, c.unread_count, ct.username
            FROM conversations c
            LEFT JOIN contacts ct ON ct.peer_id = c.peer_id
            ORDER BY c.last_message_at DESC
            """
        ) as cursor:
            rows = await cursor.fetchall()
            return [dict(row) for row in rows]
        
    async def mark_read(self, conversation_id: int) -> None:
        await self._db.execute(
            "UPDATE conversations SET unread_count = 0 WHERE id = ?",
            (conversation_id,),
        )
        await self._db.commit()    
        
        
    async def save_message(
        self,
        peer_id: str,
        sender_peer_id: str,
        content: str,
        is_outgoing: bool = False,
        delivered: bool = False,
        client_message_id: str | None = None
    ) -> int:
        if self._fernet:
            content = self._fernet.encrypt(content.encode('utf-8')).decode('utf-8')
        conv_id = await self.get_or_create_conversation(peer_id)
        now = _now()
        cursor = await self._db.execute(
            """
            INSERT INTO messages (conversation_id, sender_peer_id, content, sent_at, is_outgoing, delivered, client_message_id)
            VALUES (?, ?, ?, ?, ?, ?, ?)
            """,
            (conv_id, sender_peer_id, content, now, int(is_outgoing), int(delivered), client_message_id),
        )
        
        if not is_outgoing:
            await self._db.execute(
                """
                UPDATE conversations
                SET last_message_at = ?,
                    unread_count = unread_count + 1
                WHERE id = ?
                """,
                (now, conv_id),
            )
        else:
            await self._db.execute(
                "UPDATE conversations SET last_message_at = ? WHERE id = ?",
                (now, conv_id),
            )
        await self._db.commit()
        assert cursor.lastrowid is not None, "insert failed no lastrowid returned"
        return cursor.lastrowid

       # yo phone linging.. yo phone linging big boy come pick up ur phone, why you no pick up ur phone ... yo phone linging 
       # ignore these random comments ahh
       
       #tung tung tung tung tung tung tung sahur :O
    async def mark_delivered(self, client_id: str) -> None:
        await self._db.execute(
            "UPDATE messages SET delivered = 2 WHERE client_message_id = ?", (client_id,)
        )
        await self._db.commit()
    
    # ballerina cappuccina
    async def mark_relayed(self, client_id: str) -> None:
        await self._db.execute(
            "UPDATE messages SET delivered = 1 WHERE client_message_id = ? AND delivered = 0", (client_id,)
        )
        await self._db.commit()

    async def get_messages(
        self,
        peer_id: str,
        limit: int = 50 ,
    ) -> list[dict]:
        async with self._db.execute(
            """
            SELECT m.id, m.sender_peer_id, m.client_message_id, m.content, m.sent_at, m.is_outgoing, m.delivered
            FROM messages m
            JOIN conversations c ON c.id = m.conversation_id
            WHERE c.peer_id = ?
            ORDER BY m.sent_at DESC
            LIMIT ?
            """,
            (peer_id, limit),
        ) as cursor:
            rows = await cursor.fetchall()
            messages = []
            for row in reversed(list(rows)):
                msg = dict(row)
                if self._fernet:
                    try:
                        msg["content"] = self._fernet.decrypt(msg["content"].encode('utf-8')).decode('utf-8')
                    except InvalidToken:
                        pass
                messages.append(msg)
            return messages
    
    async def is_msg_dup(self, client_id: str) -> bool:
        async with self._db.execute("SELECT 1 FROM messages where client_message_id = ?", (client_id,)) as cur:
            return await cur.fetchone() is not None